import sys
import telnetlib
import time
from pathlib import Path

import click
import click_config_file
import yaml

from .herd import Herd
from .herd import logger
from .herd import set_verbose_level

# TODO:
#  - click.command shorthelp can also just be the first sentence of docstring
#  https://click.palletsprojects.com/en/8.1.x/documentation/#command-short-help
#  - document arguments in their docstring (has no help=)
#  - arguments can be configured in a dict and standardized across tools


def yamlprovider(file_path, cmd_name):
    logger.info("reading config from %s, cmd=%s", file_path, cmd_name)
    with open(file_path) as config_data:
        full_config = yaml.safe_load(config_data)
    return full_config


@click.group(context_settings={"help_option_names": ["-h", "--help"], "obj": {}})
@click.option(
    "--inventory",
    "-i",
    type=str,
    default="",
    help="List of target hosts as comma-separated string or path to ansible-style yaml file",
)
@click.option(
    "--limit",
    "-l",
    type=str,
    default="",
    help="Comma-separated list of hosts to limit execution to",
)
@click.option("--user", "-u", type=str, help="User name for login to nodes")
@click.option(
    "--key-filename",
    "-k",
    type=click.Path(exists=True, readable=True, file_okay=True, dir_okay=False),
    help="Path to private ssh key file",
)
@click.option("-v", "--verbose", count=True, default=2)
@click.pass_context
def cli(ctx, inventory, limit, user, key_filename, verbose) -> click.Context:
    """A primary set of options to configure how to interface the herd

    :param ctx:
    :param inventory:
    :param limit:
    :param user:
    :param key_filename:
    :param verbose:
    :return:
    """
    set_verbose_level(verbose)
    ctx.obj["herd"] = Herd(inventory, limit, user, key_filename)
    return ctx  # calm linter


@cli.command(short_help="Power off shepherd nodes")
@click.option("--restart", "-r", is_flag=True, help="Reboot")
@click.pass_context
def poweroff(ctx, restart):
    exit_code = ctx.obj["herd"].poweroff(restart)
    sys.exit(exit_code)


@cli.command(short_help="Run COMMAND on the shell")
@click.pass_context
@click.argument("command", type=str)
@click.option("--sudo", "-s", is_flag=True, help="Run command with sudo")
def run(ctx, command, sudo):
    replies = ctx.obj["herd"].run_cmd(sudo, command)
    for i, hostname in enumerate(ctx.obj["herd"].hostnames.values()):
        click.echo(f"\n************** {hostname} **************")
        click.echo(replies[i].stdout)
        click.echo(f"exit-code = {replies[i].exited}")
    exit_code = max([reply.exited for reply in replies.values()])
    sys.exit(exit_code)


@cli.command(short_help="Record IV data from a harvest-source")
@click.option(
    "--output_path",
    "-o",
    type=click.Path(),
    default=Herd.path_default,
    help="Dir or file path for resulting hdf5 file",
)
@click.option(
    "--algorithm",
    "-a",
    type=str,
    default=None,
    help="Choose one of the predefined virtual harvesters",
)
@click.option(
    "--duration",
    "-d",
    type=click.FLOAT,
    help="Duration of recording in seconds",
)
@click.option("--force_overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.option(
    "--use_cal_default",
    "-c",
    is_flag=True,
    help="Use default calibration values",
)
@click.option(
    "--no-start",
    "-n",
    is_flag=True,
    help="Start shepherd synchronized after uploading config",
)
@click.pass_context
def harvester(
    ctx,
    output_path,
    algorithm,
    duration,
    force_overwrite,
    use_cal_default,
    no_start,
):
    fp_output = Path(output_path)
    if not fp_output.is_absolute():
        fp_output = Herd.path_default / output_path

    parameter_dict = {
        "output_path": str(fp_output),
        "harvester": algorithm,
        "duration": duration,
        "force_overwrite": force_overwrite,
        "use_cal_default": use_cal_default,
    }

    ts_start = delay = 0
    if not no_start:
        ts_start, delay = ctx.obj["herd"].find_consensus_time()
        parameter_dict["start_time"] = ts_start

    ctx.obj["herd"].configure_measurement(
        "harvester",
        parameter_dict,
    )

    if not no_start:
        logger.info("Scheduling start of shepherd at %d (in ~ %.2f s)", ts_start, delay)
        exit_code = ctx.obj["herd"].start_measurement()
        logger.info("Shepherd started.")
        if exit_code > 0:
            logger.debug("-> max exit-code = %d", exit_code)


@cli.command(
    short_help="Emulate data, where INPUT is an hdf5 file containing harvesting data",
)
@click.argument("input_path", type=click.Path())
@click.option(
    "--output_path",
    "-o",
    type=click.Path(),
    default=Herd.path_default,
    help="Dir or file path for resulting hdf5 file with load recordings",
)
@click.option(
    "--duration",
    "-d",
    type=click.FLOAT,
    help="Duration of recording in seconds",
)
@click.option("--force_overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.option(
    "--use_cal_default",
    "-c",
    is_flag=True,
    help="Use default calibration values",
)
@click.option(
    "--enable_io/--disable_io",
    default=True,
    help="Switch the GPIO level converter to targets on/off",
)
@click.option(
    "--io_target",
    type=str,
    default="A",
    help="Choose Target that gets connected to IO",
)
@click.option(
    "--pwr_target",
    type=str,
    default="A",
    help="Choose (main)Target that gets connected to virtual Source / current-monitor",
)
@click.option(
    "--aux_voltage",
    "-x",
    type=float,
    help="Set Voltage of auxiliary Power Source (second target)",
)
@click.option(
    "--virtsource",
    "-a",  # -v & -s already taken for sheep, so keep it consistent with hrv (algorithm)
    default="direct",
    help="Use the desired setting for the virtual source",
)
@click_config_file.configuration_option(provider=yamlprovider, implicit=False)
@click.option(
    "--no-start",
    "-n",
    is_flag=True,
    help="Start shepherd synchronized after uploading config",
)
@click.pass_context
def emulator(
    ctx,
    input_path,
    output_path,
    duration,
    force_overwrite,
    use_cal_default,
    enable_io,
    io_target,
    pwr_target,
    aux_voltage,
    virtsource,
    no_start,
):
    fp_input = Path(input_path)
    if not fp_input.is_absolute():
        fp_input = Herd.path_default / input_path

    parameter_dict = {
        "input_path": str(fp_input),
        "force_overwrite": force_overwrite,
        "duration": duration,
        "use_cal_default": use_cal_default,
        "enable_io": enable_io,
        "io_target": io_target,
        "pwr_target": pwr_target,
        "aux_target_voltage": aux_voltage,
        "virtsource": virtsource,
    }

    if output_path is not None:
        fp_output = Path(output_path)
        if not fp_output.is_absolute():
            fp_output = Herd.path_default / output_path

        parameter_dict["output_path"] = str(fp_output)

    ts_start = delay = 0
    if not no_start:
        ts_start, delay = ctx.obj["herd"].find_consensus_time()
        parameter_dict["start_time"] = ts_start

    ctx.obj["herd"].configure_measurement(
        "emulator",
        parameter_dict,
    )

    if not no_start:
        logger.info("Scheduling start of shepherd at %d (in ~ %.2f s)", ts_start, delay)
        exit_code = ctx.obj["herd"].start_measurement()
        logger.info("Shepherd started.")
        if exit_code > 0:
            logger.debug("-> max exit-code = %d", exit_code)


@cli.command(
    short_help="Start pre-configured shp-service (/etc/shepherd/config.yml, UNSYNCED)",
)
@click.pass_context
def start(ctx) -> None:
    if ctx.obj["herd"].check_state():
        logger.info("Shepherd still active, will skip this command!")
        sys.exit(1)
    else:
        exit_code = ctx.obj["herd"].start_measurement()
        logger.info("Shepherd started.")
        if exit_code > 0:
            logger.debug("-> max exit-code = %d", exit_code)


@cli.command(short_help="Information about current shepherd measurement")
@click.pass_context
def check(ctx) -> None:
    if ctx.obj["herd"].check_state():
        logger.info("Shepherd still active!")
        sys.exit(1)
    else:
        logger.info("Shepherd not active! (measurement is done)")


@cli.command(short_help="Stops any harvest/emulation")
@click.pass_context
def stop(ctx) -> None:
    exit_code = ctx.obj["herd"].stop_measurement()
    logger.info("Shepherd stopped.")
    if exit_code > 0:
        logger.debug("-> max exit-code = %d", exit_code)


@cli.command(
    short_help="Uploads a file FILENAME to the remote node, stored in in REMOTE_PATH",
)
@click.argument(
    "filename",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.option(
    "--remote_path",
    "-r",
    default=Herd.path_default,
    type=click.Path(),
    help="for safety only allowed: /var/shepherd/* or /etc/shepherd/*",
)
@click.option("--force_overwrite", "-f", is_flag=True, help="Overwrite existing file")
@click.pass_context
def distribute(ctx, filename, remote_path, force_overwrite):
    ctx.obj["herd"].put_file(filename, remote_path, force_overwrite)


@cli.command(short_help="Retrieves remote hdf file FILENAME and stores in in OUTDIR")
@click.argument("filename", type=click.Path())
@click.argument(
    "outdir",
    type=click.Path(
        exists=True,
    ),
)
@click.option(
    "--timestamp",
    "-t",
    is_flag=True,
    help="Add current timestamp to measurement file",
)
@click.option(
    "--separate",
    "-s",
    is_flag=True,
    help="Every remote node gets own subdirectory",
)
@click.option(
    "--delete",
    "-d",
    is_flag=True,
    help="Delete the file from the remote filesystem after retrieval",
)
@click.option(
    "--force-stop",
    "-f",
    is_flag=True,
    help="Stop the on-going harvest/emulation process before retrieving the data",
)
@click.pass_context
def retrieve(ctx, filename, outdir, timestamp, separate, delete, force_stop) -> None:
    """

    :param ctx: context
    :param filename: remote file with absolute path or relative in '/var/shepherd/recordings/'
    :param outdir: local path to put the files in 'outdir/[node-name]/filename'
    :param timestamp:
    :param separate:
    :param delete:
    :param force_stop:
    """

    if force_stop:
        ctx.obj["herd"].stop_measurement()
        if ctx.obj["herd"].await_stop(timeout=30):
            raise Exception("shepherd still active after timeout")

    failed = ctx.obj["herd"].get_file(filename, outdir, timestamp, separate, delete)
    sys.exit(failed)


# #############################################################################
#                               OpenOCD Programmer
# #############################################################################


@cli.group(
    short_help="Remote programming/debugging of the target sensor node",
    invoke_without_command=True,
)
@click.option(
    "--port",
    "-p",
    type=int,
    default=4444,
    help="Port on which OpenOCD should listen for telnet",
)
@click.option(
    "--on/--off",
    default=True,
    help="Enable/disable power and debug access to the target",
)
@click.option("--voltage", "-v", type=float, default=3.0, help="Target supply voltage")
@click.option(
    "--sel_a/--sel_b",
    default=True,
    help="Choose (main)Target that gets connected to virtual Source",
)
@click.pass_context
def target(ctx, port, on, voltage, sel_a):
    # TODO: dirty workaround for deprecated openOCD code
    #   - also no usage of cnx.put, cnx.get, cnx.run, cnx.sudo left
    ctx.obj["openocd_telnet_port"] = port
    sel_target = "sel_a" if sel_a else "sel_b"
    if on or ctx.invoked_subcommand:
        ctx.obj["herd"].run_cmd(
            sudo=True,
            cmd=f"shepherd-sheep target-power --on --voltage {voltage} --{sel_target}",
        )
        for cnx in ctx.obj["herd"].group:
            start_openocd(cnx, ctx.obj["herd"].hostnames[cnx.host])
    else:
        replies1 = ctx.obj["herd"].run_cmd(
            sudo=True,
            cmd="systemctl stop shepherd-openocd",
        )
        replies2 = ctx.obj["herd"].run_cmd(
            sudo=True,
            cmd="shepherd-sheep target-power --off",
        )
        exit_code = max(
            [reply.exited for reply in replies1.values()]
            + [reply.exited for reply in replies2.values()],
        )
        sys.exit(exit_code)


# @target.result_callback()  # TODO: disabled for now: errors in recent click-versions
@click.pass_context
def process_result(ctx, result, **kwargs):
    if not kwargs["on"]:
        replies1 = ctx.obj["herd"].run_cmd(
            sudo=True,
            cmd="systemctl stop shepherd-openocd",
        )
        replies2 = ctx.obj["herd"].run_cmd(
            sudo=True,
            cmd="shepherd-sheep target-power --off",
        )
        exit_code = max(
            [reply.exited for reply in replies1.values()]
            + [reply.exited for reply in replies2.values()],
        )
        sys.exit(exit_code)


def start_openocd(cnx, hostname, timeout=30):
    # TODO: why start a whole telnet-session? we can just flash and verify firmware by remote-cmd
    # TODO: bad design for parallelization, but deprecated anyway
    cnx.sudo("systemctl start shepherd-openocd", hide=True, warn=True)
    ts_end = time.time() + timeout
    while True:
        openocd_status = cnx.sudo(
            "systemctl status shepherd-openocd",
            hide=True,
            warn=True,
        )
        if openocd_status.exited == 0:
            break
        if time.time() > ts_end:
            raise TimeoutError(f"Timed out waiting for openocd on host {hostname}")
        else:
            logger.debug("waiting for openocd on %s", hostname)
            time.sleep(1)


@target.command(short_help="Flashes the binary IMAGE file to the target")
@click.argument(
    "image",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.pass_context
def flash(ctx, image):
    for cnx in ctx.obj["herd"].group:
        hostname = ctx.obj["herd"].hostnames[cnx.host]
        cnx.put(image, "/tmp/target_image.bin")  # noqa: S108

        with telnetlib.Telnet(cnx.host, ctx.obj["openocd_telnet_port"]) as tn:
            logger.debug("connected to openocd on %s", hostname)
            tn.write(b"program /tmp/target_image.bin verify reset\n")
            res = tn.read_until(b"Verified OK", timeout=5)
            if b"Verified OK" in res:
                logger.info("flashed image on %s successfully", hostname)
            else:
                logger.error("failed flashing image on %s", hostname)


@target.command(short_help="Halts the target")
@click.pass_context
def halt(ctx):
    for cnx in ctx.obj["herd"].group:
        hostname = ctx.obj["herd"].hostnames[cnx.host]

        with telnetlib.Telnet(cnx.host, ctx.obj["openocd_telnet_port"]) as tn:
            logger.debug("connected to openocd on %s", hostname)
            tn.write(b"halt\n")
            logger.info("target halted on %s", hostname)


@target.command(short_help="Erases the target")
@click.pass_context
def erase(ctx):
    for cnx in ctx.obj["herd"].group:
        hostname = ctx.obj["herd"].hostnames[cnx.host]

        with telnetlib.Telnet(cnx.host, ctx.obj["openocd_telnet_port"]) as tn:
            logger.debug("connected to openocd on %s", hostname)
            tn.write(b"halt\n")
            logger.info("target halted on %s", hostname)
            tn.write(b"nrf52 mass_erase\n")
            logger.info("target erased on %s", hostname)


@target.command(short_help="Resets the target")
@click.pass_context
def reset(ctx):
    for cnx in ctx.obj["herd"].group:
        hostname = ctx.obj["herd"].hostnames[cnx.host]
        with telnetlib.Telnet(cnx.host, ctx.obj["openocd_telnet_port"]) as tn:
            logger.debug("connected to openocd on %s", hostname)
            tn.write(b"reset\n")
            logger.info("target reset on %s", hostname)


# #############################################################################
#                               Pru Programmer
# #############################################################################


@cli.command(
    short_help="Programmer for Target-Controller",
    context_settings={"ignore_unknown_options": True},
)
@click.argument(
    "firmware-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.option(
    "--sel_a/--sel_b",
    default=True,
    help="Choose Target-Port for programming",
)
@click.option(
    "--voltage",
    "-v",
    type=click.FLOAT,
    default=3.0,
    help="Target supply voltage",
)
@click.option(
    "--speed",
    "-s",
    type=click.INT,
    default=1_000_000,
    help="Programming-Datarate",
)
@click.option(
    "--target",
    "-t",
    type=click.Choice(["nrf52", "msp430"]),
    default="nrf52",
    help="Target chip",
)
@click.pass_context
def programmer(ctx, firmware_file, sel_a, voltage, speed, target):
    temp_file = "/tmp/target_image.bin"  # noqa: S108
    ctx.obj["herd"].put_file(firmware_file, temp_file, force_overwrite=True)
    command = (
        f"shepherd-sheep programmer {temp_file} --sel_{'a' if sel_a else 'b'} "
        f"-v {voltage} -s {speed} -t {target}"
    )
    replies = ctx.obj["herd"].run_cmd(sudo=True, cmd=command)
    exit_code = max([reply.exited for reply in replies.values()])
    sys.exit(exit_code)


if __name__ == "__main__":
    cli()



import time

import pytest
from shepherd_herd.cli import cli

from .conftest import wait_for_end


@pytest.mark.timeout(120)
def test_hrv_example(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "harvester",
            "-a",
            "cv20",
            "-d",
            "10",
            "-o",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(60)
def test_hrv_example_fail(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "harvester",
            "--algorithm",
            "ceeeveeeee",
            "--duration",
            "10",
            "--output_path",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=15)


@pytest.mark.timeout(60)
def test_hrv_minimal(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        ["harvester"],
    )
    assert res.exit_code == 0
    time.sleep(10)
    # forced stop
    res = cli_runner.invoke(
        cli,
        ["-vvv", "stop"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=10)


@pytest.mark.timeout(120)
def test_hrv_all_args_long(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "harvester",
            "--algorithm",
            "cv33",
            "--duration",
            "10",
            "--force_overwrite",
            "--use_cal_default",
            "--output_path",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(120)
def test_hrv_all_args_short(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "harvester",
            "-a",
            "cv33",
            "-d",
            "10",
            "-f",
            "-c",
            "-o",
            "pytest_hrv.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
def test_hrv_no_start(cli_runner, stopped_herd) -> None:
    # Note: short timeout is the catch
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "harvester",
            "-d",
            "10",
            "--no-start",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=10)
    # manual start
    res = cli_runner.invoke(
        cli,
        ["-vvv", "start"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


# TODO: retrieve & check with datalib (length & validity)



import time
from pathlib import Path
from shutil import copy

import numpy as np
import pytest
import yaml
from shepherd_data import Writer
from shepherd_herd.cli import cli


def extract_first_sheep(herd_path: Path) -> str:
    with open(herd_path) as stream:
        try:
            inventory_data = yaml.safe_load(stream)
        except yaml.YAMLError:
            raise TypeError(f"Couldn't read inventory file {herd_path}")
    return list(inventory_data["sheep"]["hosts"].keys())[0]


def wait_for_end(cli_run, tmin: float = 0, timeout: float = 999) -> bool:
    ts_start = time.time()
    while cli_run.invoke(cli, ["-vvv", "check"]).exit_code > 0:
        duration = time.time() - ts_start
        if duration > timeout:
            raise TimeoutError(f"Shepherd ran into timeout ({timeout} s)")
        time.sleep(2)
    duration = time.time() - ts_start
    if duration < tmin:
        raise TimeoutError(f"Shepherd only took {duration} s (min = {tmin} s)")
    return False


def generate_h5_file(file_path: Path, file_name: str = "harvest_example.h5") -> Path:
    store_path = file_path / file_name

    with Writer(store_path, compression=0) as file:
        file.set_hostname("artificial")
        duration_s = 2
        repetitions = 10
        timestamp_vector = np.arange(0.0, duration_s, file.sample_interval_ns / 1e9)

        # values in SI units
        voltages = np.linspace(3.30, 1.90, int(file.samplerate_sps * duration_s))
        currents = np.linspace(100e-6, 2000e-6, int(file.samplerate_sps * duration_s))

        for idx in range(repetitions):
            timestamps = idx * duration_s + timestamp_vector
            file.append_iv_data_si(timestamps, voltages, currents)

    return store_path


@pytest.fixture
def data_h5_path(tmp_path) -> Path:
    return generate_h5_file(tmp_path)


@pytest.fixture
def local_herd(tmp_path) -> Path:
    # locations copied from herd.cli()
    inventories = [
        "/etc/shepherd/herd.yml",
        "~/herd.yml",
        "inventory/herd.yml",
    ]
    host_path = None
    for inventory in inventories:
        if Path(inventory).exists():
            host_path = Path(inventory)
    if host_path is None:
        raise FileNotFoundError(", ".join(inventories))
    local_path = tmp_path / "herd.yml"
    copy(host_path, local_path)

    return local_path


@pytest.fixture
def stopped_herd(cli_runner):
    cli_runner.invoke(cli, ["-vvv", "stop"])
    wait_for_end(cli_runner)



import time

import pytest
from shepherd_herd.cli import cli

from .conftest import generate_h5_file
from .conftest import wait_for_end


@pytest.mark.timeout(60)
def test_emu_prepare(cli_runner, stopped_herd, tmp_path) -> None:
    # distribute file and emulate from it in following tests
    test_file = generate_h5_file(tmp_path, "pytest_src.h5")
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "distribute",
            str(test_file),
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner)


@pytest.mark.timeout(120)
def test_emu_example(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "emulator",
            "--virtsource",
            "BQ25504",
            "-o",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=20)


@pytest.mark.timeout(60)
def test_emu_example_fail(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "emulator",
            "--virtsource",
            "BQ25504",
            "-o",
            "pytest_emu.h5",
            "pytest_NonExisting.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=15)


@pytest.mark.timeout(120)
def test_emu_minimal(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "emulator",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=20)


@pytest.mark.timeout(120)
def test_emu_all_args_long(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "emulator",
            "--duration",
            "10",
            "--force_overwrite",
            "--use_cal_default",
            "--enable_io",
            "--io_target",
            "A",
            "--pwr_target",
            "A",
            "--aux_voltage",
            "1.6",
            "--virtsource",
            "BQ25504",
            "--output_path",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(120)
def test_emu_all_args_short(cli_runner, stopped_herd) -> None:
    # short arg or opposite bool val
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "emulator",
            "-d",
            "10",
            "-f",
            "-c",
            "--disable_io",
            "--io_target",
            "B",
            "--pwr_target",
            "B",
            "-x",
            "1.4",
            "-a",
            "BQ25570",
            "-o",
            "pytest_emu.h5",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(150)
def test_emu_no_start(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "emulator",
            "-d",
            "20",
            "-o",
            "pytest_emu.h5",
            "--no-start",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=15)
    # manual start
    res = cli_runner.invoke(
        cli,
        ["-vvv", "start"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, tmin=15)


@pytest.mark.timeout(60)
def test_emu_force_stop(cli_runner, stopped_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "emulator",
            "pytest_src.h5",
        ],
    )
    assert res.exit_code == 0
    time.sleep(10)
    # forced stop
    res = cli_runner.invoke(
        cli,
        ["-vvv", "stop"],
    )
    assert res.exit_code == 0
    wait_for_end(cli_runner, timeout=10)


# TODO: retrieve & check with datalib



"""
shepherd_herd
~~~~~
click-based command line utility for controlling a group of shepherd nodes
remotely through ssh. Provides commands for starting/stopping harvester and
emulator, retrieving recordings to the local machine and flashing firmware
images to target sensor nodes.

:copyright: (c) 2019 Networked Embedded Systems Lab, TU Dresden.
:license: MIT, see LICENSE for more details.
"""
from .herd import Herd
from .herd import logger
from .herd import set_verbose_level

__version__ = "0.4.3"

__all__ = [
    "Herd",
    "logger",
    "set_verbose_level",
]



"""
Herd-Baseclass
"""
import contextlib
import logging
import threading
import time
from io import StringIO
from pathlib import Path
from typing import List

import yaml
from fabric import Group

consoleHandler = logging.StreamHandler()
logger = logging.getLogger("shepherd-herd")
logger.addHandler(consoleHandler)
verbose_level = 0
# Note: defined here to avoid circular import


def set_verbose_level(verbose: int = 2) -> None:
    if verbose == 0:
        logger.setLevel(logging.ERROR)
    elif verbose == 1:
        logger.setLevel(logging.WARNING)
    elif verbose == 2:
        logger.setLevel(logging.INFO)
    elif verbose > 2:
        logger.setLevel(logging.DEBUG)
    global verbose_level
    verbose_level = verbose


class Herd:
    group: Group = None
    hostnames: dict = None

    _remote_paths_allowed = [
        Path("/var/shepherd/recordings/"),  # default
        Path("/var/shepherd/"),
        Path("/etc/shepherd/"),
        Path("/tmp/"),  # noqa: S108
    ]
    path_default = _remote_paths_allowed[0]

    timestamp_diff_allowed = 10
    start_delay_s = 20

    def __init__(
        self,
        inventory: str = "",
        limit: str = "",
        user=None,
        key_filename=None,
    ):
        if limit.rstrip().endswith(","):
            limit = limit.split(",")[:-1]
        else:
            limit = None

        if inventory.rstrip().endswith(","):
            hostlist = inventory.split(",")[:-1]
            if limit is not None:
                hostlist = list(set(hostlist) & set(limit))
            hostnames = {hostname: hostname for hostname in hostlist}

        else:
            # look at all these directories for inventory-file
            if inventory == "":
                inventories = [
                    "/etc/shepherd/herd.yml",
                    "~/herd.yml",
                    "inventory/herd.yml",
                ]
            else:
                inventories = [inventory]
            host_path = None
            for inventory in inventories:
                if Path(inventory).exists():
                    host_path = Path(inventory)

            if host_path is None:
                raise FileNotFoundError(", ".join(inventories))

            with open(host_path) as stream:
                try:
                    inventory_data = yaml.safe_load(stream)
                except yaml.YAMLError:
                    raise FileNotFoundError(f"Couldn't read inventory file {host_path}")

            hostlist = []
            hostnames = {}
            for hostname, hostvars in inventory_data["sheep"]["hosts"].items():
                if isinstance(limit, List) and (hostname not in limit):
                    continue

                if "ansible_host" in hostvars:
                    hostlist.append(hostvars["ansible_host"])
                    hostnames[hostvars["ansible_host"]] = hostname
                else:
                    hostlist.append(hostname)
                    hostnames[hostname] = hostname

            if user is None:
                with contextlib.suppress(KeyError):
                    user = inventory_data["sheep"]["vars"]["ansible_user"]

        if user is None:
            raise ValueError("Provide user by command line or in inventory file")

        if len(hostlist) < 1 or len(hostnames) < 1:
            raise ValueError(
                "Provide remote hosts (either inventory empty or limit does not match)",
            )

        connect_kwargs = {}
        if key_filename is not None:
            connect_kwargs["key_filename"] = key_filename

        self.group = Group(*hostlist, user=user, connect_kwargs=connect_kwargs)
        self.hostnames = hostnames
        logger.debug("Sheep-Herd consists of %d nodes", len(self.group))
        if len(self.group) < 1:
            raise ValueError("No remote targets found in list!")

    def __del__(self):
        # ... overcautious
        with contextlib.suppress(TypeError):
            for cnx in self.group:
                cnx.close()
                del cnx

    def __getitem__(self, key):
        if key in self.hostnames:
            return self.hostnames[key]
        raise KeyError

    def __repr__(self):
        return self.hostnames

    @staticmethod
    def thread_run(cnx, sudo: bool, cmd: str, results: dict, index: int) -> None:
        if sudo:
            results[index] = cnx.sudo(cmd, warn=True, hide=True)
        else:
            results[index] = cnx.run(cmd, warn=True, hide=True)

    def run_cmd(self, sudo: bool, cmd: str) -> dict:
        results = {}
        threads = {}
        for i, cnx in enumerate(self.group):
            threads[i] = threading.Thread(
                target=self.thread_run,
                args=(cnx, sudo, cmd, results, i),
            )
            threads[i].start()
        for thread in threads.values():
            thread.join()
            del thread  # ... overcautious
        return results

    @staticmethod
    def thread_put(cnx, src: [Path, StringIO], dst: Path, force_overwrite: bool):
        if isinstance(src, StringIO):
            filename = dst.name
        else:
            filename = src.name
            src = str(src)

        if dst.suffix == "" and not str(dst).endswith("/"):
            dst = str(dst) + "/"

        tmp_path = Path("/tmp") / filename  # noqa: S108
        logger.debug("TMP path for %s is %s", cnx.host, tmp_path)

        cnx.put(src, str(tmp_path))  # noqa: S108
        xtr_arg = "-f" if force_overwrite else "-n"
        cnx.sudo(f"mv {xtr_arg} {tmp_path} {dst}", warn=True, hide=True)

    def put_file(
        self,
        src: [StringIO, Path, str],
        dst: [Path, str],
        force_overwrite: bool = False,
    ) -> None:
        if isinstance(src, StringIO):
            src_path = src
        else:
            src_path = Path(src).absolute()
            if not src_path.exists():
                raise FileNotFoundError(
                    "Local source file '%s' does not exist!",
                    src_path,
                )
            logger.info("Local source path = %s", src_path)

        if dst is None:
            dst_path = self.path_default
            logger.debug("Remote path not provided -> default = %s", dst_path)
        else:
            dst_path = Path(dst).absolute()
            is_allowed = False
            for path_allowed in self._remote_paths_allowed:
                if str(dst_path).startswith(str(path_allowed)):
                    is_allowed = True
            if is_allowed:
                logger.info("Remote dest path = %s", dst_path)
            else:
                raise NameError(f"provided path was forbidden ('{dst_path}')")

        threads = {}
        for i, cnx in enumerate(self.group):
            threads[i] = threading.Thread(
                target=self.thread_put,
                args=(cnx, src_path, dst_path, force_overwrite),
            )
            threads[i].start()
        for thread in threads.values():
            thread.join()
            del thread  # ... overcautious

    @staticmethod
    def thread_get(cnx, src: Path, dst: Path):
        cnx.get(str(src), local=str(dst))

    def get_file(
        self,
        src: [Path, str],
        dst_dir: [Path, str],
        timestamp: bool = False,
        separate: bool = False,
        delete_src: bool = False,
    ) -> bool:
        time_str = time.strftime("%Y_%m_%dT%H_%M_%S")
        xtra_ts = f"_{time_str}" if timestamp else ""
        failed_retrieval = False

        threads = {}
        dst_paths = {}

        # assemble file-names
        if Path(src).is_absolute():
            src_path = Path(src)
        else:
            src_path = Path(self.path_default) / src

        for i, cnx in enumerate(self.group):
            hostname = self.hostnames[cnx.host]
            if separate:
                target_path = Path(dst_dir) / hostname
                xtra_node = ""
            else:
                target_path = Path(dst_dir)
                xtra_node = f"_{hostname}"

            dst_paths[i] = target_path / (
                str(src_path.stem) + xtra_ts + xtra_node + src_path.suffix
            )

        # check if file is present
        replies = self.run_cmd(sudo=False, cmd=f"test -f {src_path}")

        # try to fetch data
        for i, cnx in enumerate(self.group):
            hostname = self.hostnames[cnx.host]
            if replies[i].exited > 0:
                logger.error(
                    "remote file '%s' does not exist on node %s",
                    src_path,
                    hostname,
                )
                failed_retrieval = True
                continue

            if not dst_paths[i].parent.exists():
                logger.info("creating local dir of %s", dst_paths[i])
                dst_paths[i].parent.mkdir()

            logger.debug(
                "retrieving remote src-file '%s' from %s to local dst '%s'",
                src_path,
                hostname,
                dst_paths[i],
            )

            threads[i] = threading.Thread(
                target=self.thread_get,
                args=(cnx, src_path, dst_paths[i]),
            )
            threads[i].start()

        for i, cnx in enumerate(self.group):
            hostname = self.hostnames[cnx.host]
            if replies[i].exited > 0:
                continue
            threads[i].join()
            del threads[i]  # ... overcautious
            if delete_src:
                logger.info(
                    "deleting %s from remote %s",
                    src_path,
                    hostname,
                )
                cnx.sudo(f"rm {src_path}", hide=True)

        del threads
        return failed_retrieval

    def find_consensus_time(self) -> (int, float):
        """Finds a start time in the future when all nodes should start service

        In order to run synchronously, all nodes should start at the same time.
        This is achieved by querying all nodes to check any large time offset,
        agreeing on a common time in the future and waiting for that time on each
        node.
        """
        # Get the current time on each target node
        replies = self.run_cmd(sudo=False, cmd="date +%s")
        ts_nows = [float(reply.stdout) for reply in replies.values()]
        ts_max = max(ts_nows)
        ts_min = min(ts_nows)
        ts_diff = ts_max - ts_min
        # Check for excessive time difference among nodes
        if ts_diff > self.timestamp_diff_allowed:
            raise Exception(
                f"Time difference between hosts greater {self.timestamp_diff_allowed} s",
            )

        # We need to estimate a future point in time such that all nodes are ready
        ts_start = ts_max + self.start_delay_s
        return int(ts_start), float(self.start_delay_s + ts_diff / 2)

    def configure_measurement(
        self,
        mode: str,
        parameters: dict,
    ) -> None:
        """Configures shepherd service on the group of hosts.

        Rolls out a configuration file according to the given command and parameters
        service.

        Args:
            mode (str): What shepherd is supposed to do. One of 'harvester' or 'emulator'.
            parameters (dict): Parameters for shepherd-sheep
        """
        global verbose_level
        config_dict = {
            "mode": mode,
            "verbose": verbose_level,
            "parameters": parameters,
        }
        config_yml = yaml.dump(config_dict, default_flow_style=False, sort_keys=False)

        logger.debug("Rolling out the following config:\n\n%s", config_yml)

        if self.check_state(warn=True):
            raise Exception("shepherd still active!")

        self.put_file(
            StringIO(config_yml),
            "/etc/shepherd/config.yml",
            force_overwrite=True,
        )

    def check_state(self, warn: bool = False) -> bool:
        """Returns true ss long as one instance is still measuring

        :param warn:
        :return: True is one node is still active
        """
        replies = self.run_cmd(sudo=True, cmd="systemctl status shepherd")
        active = False

        for i, cnx in enumerate(self.group):
            if replies[i].exited != 3:
                active = True
                if warn:
                    logger.warning(
                        "shepherd still active on %s",
                        self.hostnames[cnx.host],
                    )
                else:
                    logger.debug(
                        "shepherd still active on %s",
                        self.hostnames[cnx.host],
                    )
        return active

    def start_measurement(self) -> int:
        """Starts shepherd service on the group of hosts."""
        if self.check_state(warn=True):
            logger.info("-> won't start while shepherd-instances are active")
            return 1
        else:
            replies = self.run_cmd(sudo=True, cmd="systemctl start shepherd")
            return max([reply.exited for reply in replies.values()])

    def stop_measurement(self) -> int:
        logger.debug("Shepherd-nodes affected: %s", self.hostnames.values())
        replies = self.run_cmd(sudo=True, cmd="systemctl stop shepherd")
        exit_code = max([reply.exited for reply in replies.values()])
        logger.info("Shepherd was forcefully stopped")
        if exit_code > 0:
            logger.debug("-> max exit-code = %d", exit_code)
        return exit_code

    def poweroff(self, restart: bool) -> bool:
        logger.debug("Shepherd-nodes affected: %s", self.hostnames.values())
        if restart:
            replies = self.run_cmd(sudo=True, cmd="reboot")
            logger.info("Command for rebooting nodes was issued")
        else:
            replies = self.run_cmd(sudo=True, cmd="poweroff")
            logger.info("Command for powering off nodes was issued")
        exit_code = max([reply.exited for reply in replies.values()])
        return exit_code

    def await_stop(self, timeout: int = 30) -> bool:
        ts_end = time.time() + timeout
        while self.check_state():
            if time.time() > ts_end:
                return self.check_state(warn=True)
            time.sleep(1)
        return False



import os

import pytest
from shepherd_herd.cli import cli

from .conftest import extract_first_sheep
from .conftest import generate_h5_file


@pytest.mark.timeout(10)
def test_run_standard(cli_runner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_run_extra(cli_runner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_run_fail(cli_runner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "runnnn",
            "date",
        ],
    )
    assert res.exit_code != 0


@pytest.mark.timeout(10)
def test_run_sudo(cli_runner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "run",
            "-s",
            "echo 'it's me: $USER",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_run_sudo_long(cli_runner) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "run",
            "--sudo",
            "echo 'it's me: $USER",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_inventory(cli_runner, local_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            str(local_herd),
            "-vvv",
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_inventory_long(cli_runner, local_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "--inventory",
            str(local_herd),
            "--verbose",
            "=3",
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_limit(cli_runner, local_herd) -> None:
    sheep = extract_first_sheep(local_herd)
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            str(local_herd),
            "-l",
            f"{sheep},",
            "-vvv",
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_limit_long(cli_runner, local_herd) -> None:
    sheep = extract_first_sheep(local_herd)
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            str(local_herd),
            "--limit",
            f"{sheep},",
            "-vvv",
            "run",
            "date",
        ],
    )
    assert res.exit_code == 0


@pytest.mark.timeout(10)
def test_provide_limit_fail(cli_runner, local_herd) -> None:
    res = cli_runner.invoke(
        cli,
        [
            "-i",
            str(local_herd),
            "-l",
            "MrMeeseeks,",
            "-vvv",
            "run",
            "date",
        ],
    )
    assert res.exit_code != 0


def test_distribute_retrieve_std(cli_runner, tmp_path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(os.listdir(tmp_path))
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "distribute",
            str(test_file),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "-f",
            "-t",
            "-d",
            str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(os.listdir(tmp_path))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "-s",
            str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(os.listdir(tmp_path))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


def test_distribute_retrieve_etc(cli_runner, tmp_path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(os.listdir(tmp_path))
    dir_remote = "/etc/shepherd/"
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "distribute",
            "--remote_path",
            dir_remote,
            str(test_file),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "--force-stop",
            "--separate",
            "--delete",
            dir_remote + str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(os.listdir(tmp_path))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "--timestamp",
            dir_remote + str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(os.listdir(tmp_path))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


def test_distribute_retrieve_var(cli_runner, tmp_path) -> None:
    test_file = generate_h5_file(tmp_path, "pytest_deploy.h5")
    elem_count1 = len(os.listdir(tmp_path))
    dir_remote = "/var/shepherd/"
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "distribute",
            "-r",
            dir_remote,
            str(test_file),
        ],
    )
    assert res.exit_code == 0
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "--force-stop",
            "--separate",
            "--delete",
            dir_remote + str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code == 0
    elem_count2 = len(os.listdir(tmp_path))
    # file got deleted in prev retrieve, so fail now
    res = cli_runner.invoke(
        cli,
        [
            "-vvv",
            "retrieve",
            "--timestamp",
            dir_remote + str(test_file.name),
            str(tmp_path),
        ],
    )
    assert res.exit_code != 0
    elem_count3 = len(os.listdir(tmp_path))
    assert elem_count1 < elem_count2
    assert elem_count2 == elem_count3


# TODO: test providing user and key filename
# TODO: test poweroff (reboot)
# TODO: test sudo
